#include <stddef.h>
#include <sys/mman.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "./size.h"
#include "block.h"
#include "raw_chunk_manager.h"
#include "chunk_manager.h"
#include "malloc.h"

void *alloc_tiny(size_t size)
{
	(void)size;
	return NULL;
}

void *alloc_small(size_t size)
{
	(void)size;
	return NULL;
}

void *alloc_large(size_t size)
{
	chunk_t new_chunk;
	void    *first = allocs_tree[LARGE];

	if (first == NULL)
	{
		if (add_new_block(&new_chunk, size, NULL, NULL, true))
			return NULL;
		allocs_tree[LARGE] = new_chunk.current;
		return new_chunk.data_start;
	}

	void    *last = raw_get_last_chunk(first);
	
	if (add_new_block(&new_chunk, size, last, NULL, true))
		return NULL;
	
	return new_chunk.data_start;
}

void *ft_malloc(size_t size)
{
	if (size < n)
		return alloc_tiny(size);
	if (size < m)
		return alloc_small(size);
	return alloc_large(size); 
	//mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
}

void free_tiny(chunk_t *chunk)
{
	(void) chunk;
}

void free_small(chunk_t *chunk)
{
	(void) chunk;
}

void free_large(chunk_t *chunk)
{
	if (chunk->current == allocs_tree[LARGE])
		allocs_tree[LARGE] = chunk->next;
	destroy_chunk(chunk);
	munmap(chunk->current, get_physical_size(chunk));
}

static void (*free_funcs[3])(chunk_t*) = {free_tiny, free_small, free_large};

void ft_free(void *ptr)
{
	void    **root;
	void    *raw_chunk;
	chunk_t chunk;
	size_t i;

	for (i = TINY; i <= LARGE + 1; i++)
	{
		if (i > LARGE)
		{
			// THROW ERROR
			write(2, "block not found\n", 16);
		}

		root = allocs_tree[i];
		
		raw_chunk = raw_get_chunk(root, ptr);
		if (raw_chunk == NULL)
			continue;

		break;
	}

	chunk_read(raw_chunk, &chunk);

	if (chunk.is_used == false)
	{
		write(2, "double free\n", 12);
		// THROW ERROR
	}

	// CALL THE RIGHT FREE FUNCTION DEPEND ON SIZE
	free_funcs[i](&chunk);	
}